package hiit;

import hiit.parser.BetelgeuseProtocolSocket;
import hiit.parser.ContextParser;
import hiit.parser.rule.BetelgeuseParserRule;
import hiit.transceiver.*;
import hiit.util.DbConnectivity;

import java.io.*;
import java.net.*;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Stack;
import java.util.Vector;

import java.util.Hashtable;

/**
 * BetelgeuseListener: receives data via HTTP socket connections and saves 
 * it to the BeTelGeuse database. Handles multiple socket connections via threading.
 * @author Joonas Kukkonen and Eemil Lagerspetz
 *
 */

/*
 * New version for the new database
 */

public class ContextListener {
	
	//private int port = 10001;
	private boolean test = false;
	private ServerSocket ss = null;
	//private Socket client = null;
	private boolean isRunning = true;
	
	private Hashtable<Integer, ContextParser> sessionThread = null;
	private Stack<Integer> available = null;
	private Connection conn;
	InputStream in = null;
	
	
	/**
	 * Used to start the BTG listener.
	 * @param port The port to listen on.
	 */
	public ContextListener(int port, boolean test){ 
		//this.port = port;
		this.test = test;
		this.sessionThread = new Hashtable<Integer, ContextParser>();
		
		this.available = new Stack<Integer>();
		for (int i = 65535; i > 0; i--){
			available.push(i);
		}	
		
		try {
			ss = new ServerSocket(port);
		} catch (IOException e) {
			System.err.println("Unable to open server socket. "+ e);
		}
	} 
	
	/**
	 * Starts listening on the specified port. Creates BetelgeuseProtocolSocket 
	 * instances to handle incoming connections.
	 *
	 */
	public void listen(){
		
		OutputStream out = null;
		Socket s = null;
		String bta, friendlyName, userName, password;
		int compression, senderType;
		int tempLength;
		Thread t = null;
		long dbSession = 0;
		
		while(isRunning){
			try {
				s = ss.accept();
				in = s.getInputStream();
				int[] head = new int[6]; // 1 Command 3 Length 2 SessionId
				out = s.getOutputStream();
				//Read the message and close the stream
				int num = read(head, 0, 6);
				if (num != 6){
					System.err.println("Message header not found! "+ Arrays.toString(head));
					continue;
				}
				int command = head[0];
				int length = head[1]*256*256 + head[2]*256 + head[3] -2; // sessionID already read
				if (length <= 0){
					continue;
				}
				int btsession = head[4]*256 + head[5];
				
				System.out.print("RECV: " + (char)command + " ");
				System.out.println(length + " " + btsession + " ");
				
				int offset = 0;
				
				int[] fish = new int[length];
				num =  read(fish, offset, length);

				if (num != length){
					System.err.println("Message body too short: "+ Arrays.toString(fish) + " length < "+ (length));
					continue;
				}
				
				if(!(command == 'c' || command == 'C')){
					System.err.println("Invalid command byte: "+ (char)command);
					continue;
				}
				
				//Parse the fields
				senderType = fish[offset];
				offset++;
				compression = fish[offset];
				offset++;
				bta = new String(fish, offset, 12);
				offset += 12;
				bta = bta.toLowerCase();
				tempLength = fish[offset];
				offset++;
				friendlyName = new String(fish, offset, tempLength);
				offset += tempLength;
				tempLength = fish[offset];
				offset++;
				userName = new String(fish, offset, tempLength);
				offset += tempLength;
				tempLength = fish[offset];
				password = new String(fish, offset, tempLength);
				
				//Create a Transceiver...
				
				Transceiver trans = null;
				
				switch(compression){
					case 0:	trans = new ByteStreamTransceiver(s, in, out);
							break;
				}
				
				ContextParser parser = null;
				
				switch(senderType){
					case 0: //Default BeTelGeuse parser...
						
						//FIXME
						//At the moment password is not checked and a new user is created each time a new user name is encountered...
						//In the future: check the user name and the password. If they do not mach, send an error message back to BeTelGeuse.
						//Before this, however, implement some easy way to add users to the database.

						long userId = -1;
						
						this.conn = DbConnectivity.getConnection();
						Statement st = null;
						int success = -1;
						try {
						    st = conn.createStatement();
						} catch (SQLException e) {
						    System.err.println("Unable to create statement. "+ e);
						    continue;
						}
						try{
						    ResultSet rs = st.executeQuery("select id from user where name = '" + userName + "';");
						    if(rs.first()){
							userId = rs.getLong("id");
						    }else{
							System.out.println("User not found, creating a new one.");
							success = st.executeUpdate("insert into user values(0, '', '" + userName + "', '', 0, '');", Statement.RETURN_GENERATED_KEYS);
							if(success != 1){
							    System.out.println("Could not add a new user. Failure code: " + success);
							    continue;
							}
							rs = st.getGeneratedKeys();
							if(rs.first()){
							    userId = rs.getLong(1);
							}else{
							    System.out.println("Could not insert new row into user.");
							    continue;
							}
						    }
						}catch (SQLException e){
						    System.err.println("Error while accessing the user table: " + e);
						    e.printStackTrace();
						    continue;
						}
						
						if (btsession != 0){
							parser = (BetelgeuseProtocolSocket)this.sessionThread.get(new Integer(btsession));
							if (parser != null && parser.getBluetoothAddress().equalsIgnoreCase(bta) && parser.alive()){
								
								dbSession = parser.getDBSession();
	                            //Transmit the session ID to Betelgeuse
	                            ByteArrayOutputStream byteout = new ByteArrayOutputStream();
	                            ByteArrayOutputStream byteout2 = new ByteArrayOutputStream();
	                            byteout.write("R".getBytes());
	                            
	                            //write the db session (save the length for later use...)
	                            int dbslength = longToBytes(dbSession, byteout2);
	                            
	                            //New code: send current timestamp to synchronize time with BTG
	                            String time = "" + System.currentTimeMillis();
	                            intToBytes(time.length(),1,byteout2);
	                            byteout2.write(time.getBytes());
	                            
	                            int lth = byteout2.toByteArray().length;
	                            
	                            intToBytes(4 + lth, 3, byteout);
	                            intToBytes(btsession, 2, byteout);
	                            byteout.write(0);
	                            
	                            //Write the DB session id length
	                            intToBytes(dbslength, 1, byteout);
	                            
	                            System.out.println("Sending reply: " + Arrays.toString(byteout.toByteArray()) + Arrays.toString(byteout2.toByteArray()));
	                            out.write(byteout.toByteArray());
	                            out.write(byteout2.toByteArray());
	                            out.flush();
	                            
	                            synchronized(parser){
                                    parser.updateTransceiver(trans);
                                }
							}else
								//Something is wrong, this one goes through next if...
								btsession = 0;
							//If new session was requested..
						}if (btsession == 0){
						    String application = "BeTelGeuse";
							String description = "BeTelGeuse data gathering application";
							long appId = -1;
							success = -1;
							try{
								ResultSet rs = st.executeQuery("select id from application where name='"+ application + "';");
								if (rs.first()){
									appId = rs.getLong("id");
								}
								else{
									System.out.println("Application not found, creating.");
									success = st.executeUpdate("insert into application values(0, '"+application+"', '"+description+"');", Statement.RETURN_GENERATED_KEYS);
									if (success != 1){
										System.out.println("Update failed: code " + success);
										continue;
									}
									rs = st.getGeneratedKeys();
									if (rs.first()){
										appId = rs.getLong(1);
									}else{
										System.err.println("Could not get application key.");
										continue;
									}
								}
							} catch (SQLException e) {
								System.err.println("Unable to get application: " + e);
								continue;
							}
							// now we have an
							// application id in appId.
							if (appId == -1){
								System.err.println("The impossible has happened. appId=" + appId);
								continue;
							}
						
							success = -1;
							try{
								ResultSet rs = st.executeQuery("select bluetooth.btaddress, bluetooth.friendlyname from bluetooth, user where bluetooth.user_id=user.id and user.id="+userId+";");
								
								success=-1;
								
								if (rs.first()){
									String dbBta = rs.getString("btaddress");
									if(dbBta.compareTo(bta) != 0){
										success = st.executeUpdate("insert into bluetooth values(0, NULL, '"+bta+"', '"+friendlyName+"', "+userId+");");
										if (success != 1){
											System.out.println("Update failed: code " + success);
											continue;
										}
									}
								}else{
									success = st.executeUpdate("insert into bluetooth values(0, NULL, '"+bta+"', '"+friendlyName+"', "+userId+");");
									if (success != 1){
										System.out.println("Update failed: code " + success);
										continue;
									}
								}						
							} catch (SQLException e) {
								System.err.println("Unable to get user: " + e);
								e.printStackTrace();
								continue;
							}
							if (userId == -1){
								System.err.println("The impossible has happened. userId=" + userId);
								continue;
							}
							System.out.println("Application id=" + appId +", User id="+ userId);
							// create session and 
							// get session

							success = -1;
							String update = "insert into session values(0, "+ appId + ", " + userId+ ", CURRENT_TIMESTAMP);";
							try{
								success = st.executeUpdate(update, Statement.RETURN_GENERATED_KEYS);
							} catch (SQLException e) {
								System.err.println("Unable to execute update: "+ update +" exception: " + e);
								continue;
							}
							if (success != 1){
								System.err.println("Could not add a new session id: " + success);
								continue;
							} 
							try {
								ResultSet res = st.getGeneratedKeys();
								if(res.first())
									dbSession = res.getLong(1);
								else {
									System.err.println("Unable to acquire session id.");
									continue;
								}
							} catch (SQLException e) {
								System.err.println("Unable to get key for session. " + e);
								continue;
							}
							
							// now everything should be in order and sessionId in dbSession != 0.
							if (dbSession == 0){
								System.err.println("The impossible has happened. dbsession=" + dbSession);
							}
								
							
							btsession = this.available.pop().intValue();
							
							
							System.out.println("[CL]: Creating a thread to handle " + s);
							//Create new BetelgeuseProtocolSocket
							parser = new BetelgeuseProtocolSocket(userName, bta, btsession, dbSession, trans);
							this.sessionThread.put(btsession, parser);
							
							
							//If Betelgeuse requests to continue old session...
							 
							
							//Transmit the session ID to Betelgeuse
							ByteArrayOutputStream byteout = new ByteArrayOutputStream();
							ByteArrayOutputStream byteout2 = new ByteArrayOutputStream();
							byteout.write("R".getBytes());
							
							//write the db session (save the length for later use...)
							int dbslength = longToBytes(dbSession, byteout2);
							
							//New code: send current timestamp to synchronize time with BTG
							String time = "" + System.currentTimeMillis();
							intToBytes(time.length(),1,byteout2);
							byteout2.write(time.getBytes());
							
							int lth = byteout2.toByteArray().length;
							
							intToBytes(4 + lth, 3, byteout);
							intToBytes(btsession, 2, byteout);
							byteout.write(0);
							
							//Write the DB session id length
							intToBytes(dbslength, 1, byteout);
							
							System.out.println("Sending reply: " + Arrays.toString(byteout.toByteArray()) + Arrays.toString(byteout2.toByteArray()));
							out.write(byteout.toByteArray());
							out.write(byteout2.toByteArray());
							out.flush();
							
							//Start the Thread AFTER the responce has been writen...
							t = new Thread(parser);
							t.start();
							t = null;
						}
						break;
				}
				
				Enumeration<ContextParser> enumer = this.sessionThread.elements();
				while (enumer.hasMoreElements()){
					parser = enumer.nextElement();
					if (!parser.alive()){
						this.sessionThread.remove(new Integer(parser.getSession()));
						this.available.push(new Integer(parser.getSession()));
						parser.close();
					}
				}
				
				/*
				if(!this.unregisteredLocationThreads.isEmpty()){
					enumer = this.sessionThread.elements();
					while (enumer.hasMoreElements()){
						BetelgeuseProtocolSocket bps = (BetelgeuseProtocolSocket)enumer.nextElement();
						LocationParser lp = this.unregisteredLocationThreads.get(bps.getUserName());
						if(lp != null){
							Vector rules = bps.getRules();
							for(int i = 0; i< rules.size();i++){
								BetelgeuseParserRule bpr =((BetelgeuseParserRule)(rules.elementAt(i)));
								if(bpr.getRuleName().equals("Location Rule")){
									((LocationRule)bpr).setLocationParser(lp);
								}
							}
							this.unregisteredLocationThreads.remove(bps.getUserName());
							break;
						}
					}
				}*/
				
			} catch (IOException e) {
				System.err.println("Unable to receive messages "+ e);
				e.printStackTrace();
			}	
		}
	}
	
	/**
	 * Starts the listener on the command line. 
	 * @param args args[0] is expected to be the port number to listen to.
	 */
	public static void main(String[] args){
		int port = -1;
		boolean test = false;
		if (args.length > 0) {
			try{
				port = Integer.parseInt(args[0]);
			} catch (NumberFormatException nfe){
				System.err.println("Invalid port number: "+ args[0]);
			}
			if (args.length > 1){
			    if (args[1].equals("test")){
			        test = true;
			    }
			}
		}
		else{
			System.out.println("BetelgeuseProtocolListener, version 1.0.");
			System.out.println("Usage: java hiit.BetelgeuseProtocolListener portnumber");
			System.out.println("or");
			System.out.println("java -jar betelgeuselistener.jar portnumber");
			System.exit(1);
		}
		
		ContextListener bl = new ContextListener(port, test);
		bl.listen();
	}
	
	private void intToBytes(int number, int size, ByteArrayOutputStream out){
		byte[] array = new byte[size];
		
		for (int i = size-1; i >= 0; i--){
			array[i] = (byte)(number % 256);
			number = number / 256;
		}
		try{
			out.write(array);
		}catch (IOException e){
			System.err.println("Unable to write out the bytes to " + out);
		}
	}
	
	int read(int[] arr, int from, int lth){
		if (lth == 0)
			return 0;
		
		for (int i = from; i < lth; ++i){
			try {
				arr[i] = in.read(); 
			} catch (IOException e) {
				System.err.println("[BPS]: Unable to read "+ lth +" bytes.");
				return i+1;
			}
		}
		return lth; 
	}
	
	private int longToBytes(long number, ByteArrayOutputStream out){
		Vector<Byte> v = new Vector<Byte>();
		int lth = 0;
		while (number != 0){
			v.addElement((byte)(number % 256));
			number = number / 256;
		}
		try{
			byte[] arr = new byte[v.size()];
			lth = v.size();
			int index = 0;
			for (int i = v.size() -1; i >= 0; --i){
				arr[index] = v.elementAt(i).byteValue();
				index++;
			}			
			out.write(arr);
		}catch (IOException e){
			System.err.println("Unable to write out the bytes to " + out);
			return -1;
		}
		return lth;
	}
}
